import datetime
import uuid
from typing import Union, NamedTuple

from stix2 import Relationship, base, Identity
from stix2.base import _DomainObject, _Observable
from stix2.canonicalization.Canonicalize import canonicalize


class STIXMapperSettings(NamedTuple):
    titan_client: Union['titan_client', None] = None
    api_client: Union['ApiClient', None] = None
    # Resolve GIRs numbers into full names
    girs_names: bool = True
    # Get full reports' descriptions
    report_description: bool = True
    # Get HTML representation of reports contents and return them in a format
    # understandable for OpenCTI (base64 encoded, inside specific JSON structure)
    report_attachments_opencti: bool = False

def generate_id(
    stix_class: Union[_DomainObject, Relationship, _Observable],
    **id_contributing_properties: str,
):
    if id_contributing_properties:
        name = canonicalize(id_contributing_properties, utf8=False)
        return f"{stix_class._type}--{uuid.uuid5(base.SCO_DET_ID_NAMESPACE, name)}"
    return f"{stix_class._type}--{uuid.uuid4()}"


author_name = "Intel 471 Inc."
author_identity = Identity(
    id=generate_id(Identity, name=author_name.lower(), identity_class="organization"),
    name=author_name,
    identity_class="organization",
    created=datetime.datetime(2022, 1, 1),
    modified=datetime.datetime(2022, 1, 1),
)



import datetime
import logging
from typing import Union

from pytz import UTC
from stix2 import Bundle, Indicator, Report, TLP_AMBER, DomainName, URL, Relationship

from .. import author_identity, generate_id
from ..patterning import create_domain_pattern, create_url_pattern, create_ipv4_pattern, create_file_pattern
from ..observables import create_domain, create_url, create_ipv4, create_file
from .reports import ReportMapper
from .common import StixMapper, BaseMapper, MappingConfig

log = logging.getLogger(__name__)


@StixMapper.register("iocs", lambda x: "iocTotalCount" in x)
class IOCMapper(BaseMapper):

    mapping_configs = {
        "MaliciousURL": MappingConfig(
            patterning_mapper=create_url_pattern,
            observable_mapper=create_url,
            kwargs_extractor=lambda i: {"value": i["value"]},
            name_extractor=lambda i: i["value"],
            opencti_type="Url"
        ),
        "MaliciousDomain": MappingConfig(
            patterning_mapper=create_domain_pattern,
            observable_mapper=create_domain,
            kwargs_extractor=lambda i: {"value": i["value"].split("://")[-1]},
            name_extractor=lambda i: i["value"].split("://")[-1],
            opencti_type="Domain-Name"
        ),
        "IPAddress": MappingConfig(
            patterning_mapper=create_ipv4_pattern,
            observable_mapper=create_ipv4,
            kwargs_extractor=lambda i: {"value": i["value"]},
            name_extractor=lambda i: i["value"],
            opencti_type="IPv4-Addr"
        ),
        "MD5": MappingConfig(
            patterning_mapper=create_file_pattern,
            observable_mapper=create_file,
            kwargs_extractor=lambda i: {"md5": i["value"]},
            name_extractor=lambda i: i["value"],
            opencti_type="StixFile"
        ),
        "SHA1": MappingConfig(
            patterning_mapper=create_file_pattern,
            observable_mapper=create_file,
            kwargs_extractor=lambda i: {"sha1": i["value"]},
            name_extractor=lambda i: i["value"],
            opencti_type="StixFile"
        ),
        "SHA256": MappingConfig(
            patterning_mapper=create_file_pattern,
            observable_mapper=create_file,
            kwargs_extractor=lambda i: {"sha256": i["value"]},
            name_extractor=lambda i: i["value"],
            opencti_type="StixFile"
        )
    }

    def map(self, source: dict) -> Bundle:
        container = {}
        report_mapper = ReportMapper(self.settings)
        items = source.get("iocs") or [] if "iocTotalCount" in source else [source]
        for item in items:
            ioc_type = item["type"]
            mapping_config = self.mapping_configs.get(ioc_type)
            if not mapping_config:
                continue
            report_sources = item["links"].get("reports") or []
            valid_from = datetime.datetime.fromtimestamp(item["activeFrom"] / 1000, UTC)
            valid_until = datetime.datetime.fromtimestamp(
                item["activeTill"] / 1000, UTC
            )
            if valid_from == valid_until:
                valid_until = None

            name = mapping_config.name_extractor(item)
            kwargs = mapping_config.kwargs_extractor(item)
            stix_pattern = mapping_config.patterning_mapper(**kwargs)
            observable = mapping_config.observable_mapper(author=author_identity.id, **kwargs)
            indicator = Indicator(
                id=generate_id(Indicator, pattern=stix_pattern),
                name=name,
                pattern_type="stix",
                pattern=stix_pattern,
                indicator_types=["malicious-activity"],
                valid_from=valid_from,
                valid_until=valid_until,
                created_by_ref=author_identity,
                object_marking_refs=[TLP_AMBER],
                custom_properties={"x_opencti_main_observable_type": mapping_config.opencti_type}
            )
            r1 = Relationship(
                indicator, "based-on", observable, created_by_ref=author_identity
            )
            for stix_object in [indicator, observable, r1, author_identity, TLP_AMBER]:
                container[stix_object.id] = stix_object
            for uid, stix_object in self.map_reports(
                report_mapper, report_sources, indicator, observable, r1
            ).items():
                if isinstance(stix_object, Report) and uid in container:
                    stix_object.object_refs.extend(container[uid].object_refs)
                container[uid] = stix_object
        if container:
            bundle = Bundle(*container.values(), allow_custom=True)
            return bundle

    def map_reports(
        self,
        report_mapper,
        report_sources: list,
        indicator: Indicator,
        observable: Union[URL, DomainName],
        relationship: Relationship
    ) -> dict:
        container = {}
        for report_source in report_sources:
            container.update(
                report_mapper.map_shortened_report(
                    report_source,
                    object_refs={
                        indicator.id: indicator,
                        observable.id: observable,
                        relationship.id: relationship
                    },
                )
            )
        return container



from stix2 import URL, TLP_AMBER, IPv4Address, File, DomainName


def create_url(author: str, value: str) -> URL:
    return URL(
        value=value,
        object_marking_refs=[TLP_AMBER],
        custom_properties={"x_opencti_created_by_ref": author}
    )


def create_ipv4(value: str, author: str) -> IPv4Address:
    return IPv4Address(
        value=value,
        object_marking_refs=[TLP_AMBER],
        custom_properties={"x_opencti_created_by_ref": author}
    )


def create_domain(author: str, value: str) -> DomainName:
    return DomainName(
        value=value,
        object_marking_refs=[TLP_AMBER],
        custom_properties={"x_opencti_created_by_ref": author}
    )


def create_file(author: str, md5: str = None, sha1: str = None, sha256: str = None) -> File:
    hashes = {}
    if md5:
        hashes["md5"] = md5
    if sha1:
        hashes["sha1"] = sha1
    if sha256:
        hashes["sha256"] = sha256
    return File(
        hashes=hashes,
        object_marking_refs=[TLP_AMBER],
        custom_properties={"x_opencti_created_by_ref": author}
    )



import base64
import json
import os

import pytest
from stix2 import DomainName

from titan_client.titan_stix import STIXMapperSettings
from titan_client.titan_stix.mappers import ReportMapper
from titan_client.titan_stix.mappers.common import StixMapper

from .conftest import PREFIX, read_fixture
from mock import MagicMock

os.environ['I471_TITAN_CLIENT_CACHE_TTL'] = '0'

fixtures = {
    'test_indicators': ("indicators_input.json", "indicators_stix.json"),
    'test_iocs': ("iocs_input.json", "iocs_stix.json"),
    'test_yara': ("yara_input.json", "yara_stix.json"),
    'test_cves': ("cves_input.json", "cves_stix.json")
}


def strip_random_values(bundle: dict) -> dict:
    bundle["id"] = None
    for i1, o in enumerate(bundle["objects"]):
        bundle["objects"][i1]["created"] = None
        bundle["objects"][i1]["modified"] = None
        if o["id"].startswith("relationship--"):
            bundle["objects"][i1]["id"] = None
        for i2, object_ref_id in enumerate(o.get("object_refs", [])):
            if object_ref_id.startswith("relationship--"):
                bundle["objects"][i1]["object_refs"][i2] = None

    return bundle


@pytest.mark.parametrize('fixtures', fixtures.values(), ids=fixtures.keys())
def test_stix_mappers(fixtures):
    in_fixture, out_fixture = fixtures
    api_response = read_fixture(f'{PREFIX}/fixtures/{in_fixture}')
    expected_result = read_fixture(f'{PREFIX}/fixtures/{out_fixture}')
    mapper = StixMapper()
    result = mapper.map(api_response)
    expected = strip_random_values(expected_result)
    assert expected == strip_random_values(json.loads(result.serialize()))


def test_report_enrichments():
    api_cls = "ReportsApi"
    method_name = "reports_uid_get"
    api_response = {"rawText": "<h2>Foo</h2><p>New malware <strong>Foobar</strong> released!</p><h2>Bar</h2>"}

    api_response_mock = MagicMock(name="API response")
    api_response_mock.to_dict.return_value = api_response
    api_instance_mock = MagicMock(name="API instance")
    getattr(api_instance_mock, method_name).return_value = api_response_mock
    mock_titan_client = MagicMock(name="titan_client")
    getattr(mock_titan_client, api_cls).return_value = api_instance_mock
    mock_api_client = MagicMock()
    mock_domain = DomainName(value="foo.bar")

    mapper = ReportMapper(STIXMapperSettings(mock_titan_client, mock_api_client, report_attachments_opencti=True))
    result = mapper.map_shortened_report({
        "subject": "New malware released",
        "released": 1679321907000,
        "portalReportUrl": "https://foo.bar/fintel/123",
        "uid": "1fffffffffffffffffffffffffffffff",
        "admiraltyCode": "A1",
        "dateOfInformation": 1678060800000
    }, {mock_domain.id: mock_domain})
    report_serialized = json.loads(list(result.values())[0].serialize())
    assert report_serialized["name"] == "New malware released"
    assert report_serialized["description"] == "New malware Foobar released!"
    assert report_serialized["report_types"] == ["fintel"]
    assert report_serialized["confidence"] == 90
    assert report_serialized["x_opencti_files"] == [
        {
            "name": "Raw Text.html",
            "mime_type": "text/html",
            "data": base64.b64encode(bytes(api_response["rawText"], "utf-8")).decode("utf-8")
        }
    ]



from stix2 import Malware, TLP_AMBER
from . import generate_id, author_identity


def create_malware(value: str) -> Malware:
    return Malware(
        id=generate_id(Malware, name=value.strip().lower()),
        name=value,
        is_family=True,
        created_by_ref=author_identity,
        object_marking_refs=[TLP_AMBER],
    )



import datetime
import logging
import re
from typing import Union, NamedTuple, List, Callable
import base64

from pytz import UTC
from stix2 import Bundle, Report, ExternalReference, TLP_AMBER

from .common import BaseMapper
from .. import author_identity, generate_id, STIXMapperSettings

log = logging.getLogger(__name__)


class ReportSettings(NamedTuple):
    api_class: str
    method_name: str
    # either JSON path to the field or a function that extracts value from provided source
    description_source: Union[str, Callable[[dict], str]]
    attachments_fields: Union[List[str], None] = None


class ReportMapper(BaseMapper):
    remove_html_regex = re.compile(r"<.*?>")
    reports_settings = {
        "breach_alert": ReportSettings(
            "ReportsApi",
            "breach_alerts_uid_get",
            "data.breach_alert.summary",  # good for description
        ),
        # breach_alert example
        # titan 'iocs?ioc=akiral2iz6a7qgd3ayp3l6yub7xx2uep76idk3u2kollpj5z3z636bad.onion'
        # titan breachAlerts/fae604164acb6a34983d3684ecb83d7e
        "inforep": ReportSettings(
            "ReportsApi",
            "reports_uid_get",
            "executiveSummary",
            ["rawText", "rawTextTranslated", "researcherComments"]
        ),
        # inforep example
        # titan 'iocs?ioc=sldltcn2d6mgtp66vgmvjptdtwgqyyewsjgwkzjybq3x55plzw4tefid.onion'
        # titan reports/1d2d1de3603abf9503471b7f882c47e83e50fa42ad7588fc5d43a7cde1ff8bb4
        "spotrep": ReportSettings(
            "ReportsApi",
            "spot_reports_uid_get",
            "data.spot_report.spot_report_data.text"
        ),
        # spotrep example
        # titan 'iocs?ioc=22d90ad1a0e5220ec0772918fa6efdb54604bddab1d5f15156ead1acd5d7aa37'
        # titan spotReports/5f486c27fb68281e23aa6d2c3b0d2b59
        "fintel": ReportSettings(
            "ReportsApi",
            "reports_uid_get",
            # There's no summary or anything similar. Try to extract contents of the first paragraph <h2>...</h2><p>get-this</p>...
            lambda x: re.split(r'</?p>', re.sub(r'^.*?<p>', '', x.get("rawText") or ""))[0],
            ["rawText"]
        ),
        # fintel example
        # titan 'iocs?ioc=http://162.55.187.234'
        "malrep": ReportSettings(
            "ReportsApi",
            "malware_reports_uid_get",
            "data.malware_report_data.text",
        ),
        "sitrep": ReportSettings(
            "ReportsApi",
            "situation_reports_report_uid_get",
            "data.situation_report.text",
        ),
        "cve": ReportSettings(
            "VulnerabilitiesApi",
            "cve_reports_uid_get",
            "data.cve_report.summary",
        )
    }

    def __init__(self, settings: STIXMapperSettings):
        super().__init__(settings)
        self.cache = {}

    def map(self, source: dict) -> Bundle:
        return Bundle()

    def map_shortened_report(self, source: dict, object_refs: dict) -> dict:
        """
        Handles shortened representation of reports attached to IOCs in the response of /iocs endpoint
        {"iocs": [{"links": {"reports": [{"uid": "123", "subject": "foo", ... }]}}]}
        """
        container = {}
        if all([self.settings.titan_client, self.settings.api_client]) and \
           any([self.settings.report_description, self.settings.report_attachments_opencti]):
            try:
                full_source = self._get_full_source(source)
                if full_source:
                    # Need to merge instead of using full_source only as subjects are consistently stored
                    # under `subject` field only in shortened version
                    source = {**source, **full_source}
            except Exception as e:
                log.warning("Unable to fetch or process full report source. Falling back to shortened version. Error: %s", e)
        # TODO: researcher comments/attachments
        # TODO: Proxy setup
        report_kwargs = self._map_report_kwargs(source, object_refs)
        report = Report(**report_kwargs)
        container[report.id] = report
        return container

    def _map_report_kwargs(self, source: dict, object_refs: dict):
        time_published = self._format_published(source.get("released") or source.get("created"))
        return {
            "id": generate_id(
                Report,
                name=self.shorten(source["subject"], 128).strip().lower(),
                published=time_published,
            ),
            "name": source["subject"],
            "description": self._get_description(source),
            "report_types": [self._get_type(source)],
            "confidence": self.map_confidence(source.get("admiraltyCode") or source.get("data", {}).get("breach_alert", {}).get("confidence", {}).get("level")),
            "published": time_published,
            "labels": self._get_malware_families(source),
            "object_refs": object_refs.values(),
            "external_references": [
                ExternalReference(source_name="Titan URL", url=self._get_url(source))
            ],
            "created_by_ref": author_identity,
            "object_marking_refs": [TLP_AMBER],
            "custom_properties": {
                "x_intel471_com_uid": self._get_uid(source),
                "x_opencti_files": self._get_opencti_files(source)
            }
        }

    def _get_full_source(self, source: dict) -> Union[dict, None]:
        report_url = self._get_url(source)
        report_type = self._get_type(source)
        report_uid = self._get_uid(source)
        if report_url not in self.cache:
            report_settings = self.reports_settings.get(report_type)
            api_instance = getattr(self.settings.titan_client, report_settings.api_class)(self.settings.api_client)
            api_response = getattr(api_instance, report_settings.method_name)(report_uid)
            self.cache[report_url] = api_response.to_dict(serialize=True)
        return self.cache[report_url]

    def _format_published(self, epoch_millis: int):
        """
        Formatting datetime object for use as ID contributing property in a same way as it's done by OpenCTI
        to have the same ID here and in OpenCTI.
        """
        parsed = datetime.datetime.fromtimestamp(epoch_millis / 1000, UTC)
        return parsed.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

    @staticmethod
    def _get_url(source: dict):
        return source["portalReportUrl"]

    def _get_uid(self, source: dict):
        return self._get_url(source).split("/")[-1]

    def _get_type(self, source: dict):
        return self._get_url(source).split("/")[-2]

    def _get_malware_families(self, source: dict):
        return [i.get("value") for i in (source.get("entities") or []) if i.get("type") == "MalwareFamily"]

    def _get_description(self, source: dict):
        subject = source.get("subject")
        report_settings = self.reports_settings.get(self._get_type(source))
        description_source = report_settings.description_source

        if isinstance(description_source, Callable):
            source = description_source(source)
        else:
            for i in description_source.split("."):
                source = source.get(i, {})

        if source:
            return re.sub(self.remove_html_regex, "", source)
        return subject

    def _get_opencti_files(self, source: dict):
        if not self.settings.report_attachments_opencti:
            return []
        opencti_files = []
        report_settings = self.reports_settings.get(self._get_type(source))
        attachments_fields = report_settings.attachments_fields or []
        for field_name in attachments_fields:
            value = source.get(field_name)
            if isinstance(value, str):
                opencti_files.append({
                    "name":  " ".join([i.capitalize() for i in re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))|[a-z]+', field_name)]) + ".html",
                    "mime_type": "text/html",
                    "data": base64.b64encode(bytes(value, "utf-8")).decode("utf-8")
                })
        return opencti_files
